"""
File: FastKAN.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang; original author: Li Ziyao
Description: Modified copy of FastKAN by Li Ziyao
             (see https://github.com/ZiyaoLi/fast-kan/blob/master/fastkan/fastkan.py)
             Changes include:
                - Replaced LayerNorm with BatchNorm
                - Extended plot_curve(...) Method to work with our KANPlotter
                - modified FastKAN.__init__() to include BatchNorm and created attribute to store activations in
                - added method FastKAN.gather_activations(x), which saves all node's and edge's activations to
                  calculate relevance scores later on
                - added functions FastKAN.{get_edge_activations(), get_node_activations(),
                  get_kan_layers(), get_layersizes(), plot_curve()} to support our interface WrapperTemplate
"""
# Copyright 2024 Li, Ziyao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import *


class SplineLinear(nn.Linear):
    def __init__(self, in_features: int, out_features: int, init_scale: float = 0.1, **kw) -> None:
        self.init_scale = init_scale
        super().__init__(in_features, out_features, bias=False, **kw)

    def reset_parameters(self) -> None:
        nn.init.trunc_normal_(self.weight, mean=0, std=self.init_scale)


class RadialBasisFunction(nn.Module):
    def __init__(
            self,
            grid_min: float = -2.,
            grid_max: float = 2.,
            num_grids: int = 8,
            denominator: float = None,  # larger denominators lead to smoother basis
    ):
        super().__init__()
        self.grid_min = grid_min
        self.grid_max = grid_max
        self.num_grids = num_grids
        grid = torch.linspace(grid_min, grid_max, num_grids)
        self.grid = torch.nn.Parameter(grid, requires_grad=False)
        self.denominator = denominator or (grid_max - grid_min) / (num_grids - 1)

    def forward(self, x):
        return torch.exp(-((x[..., None] - self.grid) / self.denominator) ** 2)


class FastKANLayer(nn.Module):
    def __init__(
            self,
            input_dim: int,
            output_dim: int,
            grid_min: float = -2.,
            grid_max: float = 2.,
            num_grids: int = 8,
            use_batchnorm: bool = True,
            use_base_update: bool = True,
            base_activation=F.silu,
            spline_weight_init_scale: float = 0.1,
    ) -> None:
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.use_batchnorm = use_batchnorm
        if self.use_batchnorm:  # if BatchNorm should be used
            self.norm = nn.BatchNorm1d(output_dim, affine=True)
        self.rbf = RadialBasisFunction(grid_min, grid_max, num_grids)
        self.spline_linear = SplineLinear(input_dim * num_grids, output_dim, spline_weight_init_scale)
        self.use_base_update = use_base_update
        if use_base_update:  # SiLU function
            self.base_activation = base_activation
            # with weight and bias per edge
            self.base_linear = nn.Linear(input_dim, output_dim)

    def forward(self, x, use_layernorm=True):
        spline_basis = self.rbf(x)
        ret = self.spline_linear(spline_basis.view(*spline_basis.shape[:-2], -1))
        if self.use_base_update:
            base = self.base_linear(self.base_activation(x))
            ret = ret + base
        # normalize AFTER current layer (so normalize AFTER layer 0,1,2... but not after the last layer)
        if self.use_batchnorm:
            ret = self.norm(ret)
        return ret

    def plot_curve(
            self,
            input_index: int,
            output_index: int,
            num_pts: int = 1000,
            num_extrapolate_bins: int = 2
    ):
        '''this function returns the learned curves in a FastKANLayer.
        input_index: the selected index of the input, in [0, input_dim) .
        output_index: the selected index of the output, in [0, output_dim) .
        num_pts: num of points sampled for the curve.
        num_extrapolate_bins (N_e): num of bins extrapolating from the given grids. The curve
            will be calculate in the range of [grid_min - h * N_e, grid_max + h * N_e].
        '''
        ng = self.rbf.num_grids
        h = self.rbf.denominator
        assert input_index < self.input_dim
        assert output_index < self.output_dim
        w = self.spline_linear.weight[
            output_index, input_index * ng: (input_index + 1) * ng
            ]  # num_grids,
        x = torch.linspace(
            self.rbf.grid_min - num_extrapolate_bins * h,
            self.rbf.grid_max + num_extrapolate_bins * h,
            num_pts
        )  # num_pts, num_grids
        with (torch.no_grad()):
            y = (w * self.rbf(x.to(w.dtype)))
            # functions is tuple (result_base, result_individual_rbfs, result_all_combined)
            functions = (y, None, None)
            y = y.sum(-1)
            if self.use_base_update:
                base_linear_weight = self.base_linear.weight[output_index, input_index]
                base_value = base_linear_weight * self.base_activation(x)
                base_value += self.base_linear.bias[output_index]
                functions = (base_value, functions[0], None)
                y += base_value
            functions = (functions[0], functions[1], y)

        return x, functions


class FastKAN(nn.Module):
    def __init__(
            self,
            layers_hidden: List[int],
            grid_mins: float = -2.,
            grid_maxs: float = 2.,
            num_grids: int = 8,
            use_base_update: bool = True,
            base_activation=F.silu,
            spline_weight_init_scale: float = 0.1,
            use_batchnorm = True
    ) -> None:
        super().__init__()
        self.layers_hidden = layers_hidden
        # two lambdas to determine if we need batchnorm in current layer
        is_last_layer = lambda i: i >= len(self.layers_hidden) - 2
        # use batchnorm in every layer except the last one if use_batchnorm is True
        use_batchnorm_after_layer = lambda i: not is_last_layer(i) and use_batchnorm
        self.use_norm = use_batchnorm
        self.num_grids = num_grids
        self.layers = nn.ModuleList()
        # made grid_min and grid_max a list to specify grid intervals for each layer seperately
        if type(grid_mins) != list:
            grid_mins = [grid_mins] * (len(layers_hidden)-1)
        if type(grid_maxs) != list:
            grid_maxs = [grid_maxs] * (len(layers_hidden)-1)
        self.grid_mins = grid_mins
        self.grid_maxs = grid_maxs
        for i, (in_dim, out_dim) in enumerate(zip(layers_hidden[:-1], layers_hidden[1:])):
            new_layer = FastKANLayer(in_dim, out_dim, grid_mins[i], grid_maxs[i],
                                     num_grids, use_base_update=use_base_update,
                                     base_activation=base_activation,
                                     spline_weight_init_scale=spline_weight_init_scale,
                                     use_batchnorm=use_batchnorm_after_layer(i))
            self.layers.append(new_layer)
        # Dict of Tensors, Index l (layer) contains Tensor of [n_samples x Layer-Width]
        self.node_activations = dict()
        # Dict of Tensors, Index l (layer) contains Tensor of [n_samples x Width Layer l x Width layer l+1]
        self.edge_activations = dict()

    def forward(self, x):
        # assertion to make sure we NEVER leave the grid range of the FIRST layer, as this would indicate a problem
        # with missing normalization of the dataset
        assert torch.min(torch.amin(x, dim=0)) >= self.grid_mins[0] and torch.max(torch.amax(x, dim=0)) <= self.grid_maxs[0], "Inputs into FastKAN should be in range [-1,1]"
        for layer in self.layers:
            x = layer(x)
        return x

    def gather_activations(self, x):
        for l_idx, layer in enumerate(self.layers):
            # Node Activations (Dictionary of Tensors with index specifying layer-index)
            self.node_activations[l_idx] = x
            # Forward pass ( x = layer(x) ) but in more detail
            spline_basis = layer.rbf(x)
            # Spline_linear has dimensions [output_dim x  input_dim * num_grids], transform to [output_dim x input_dim x num_grids]
            spline_linear_pernode = layer.spline_linear.weight.view(layer.spline_linear.weight.shape[0],
                                                                    layer.input_dim, layer.rbf.num_grids)
            # multiply spline_basis (shape [N x input_dim x num_grids]) and spline_linear_pernode
            # rbf_edge_activation has shape [N x input_dim x output_dim]
            rbf_edge_activations = torch.einsum("nig,oig->nio", spline_basis, spline_linear_pernode)

            # silu_edge_activations has shape [N x output_dim]  (so just one per node)
            all_edge_activations = rbf_edge_activations
            if layer.use_base_update:
                silu_edge_activations = torch.einsum("oi,ni->nio", layer.base_linear.weight, layer.base_activation(x))
                # Bias missing, but that doesn't matter for std()
                # silu_edge_activations[:,:,] += layer.base_linear.bias
                all_edge_activations += silu_edge_activations

            self.edge_activations[l_idx] = all_edge_activations

            y = torch.einsum("nio->no", all_edge_activations)
            if layer.use_batchnorm:
                y = layer.norm(y)
            x = y  # current output becomes next input
        self.node_activations[len(self.layers)] = x  # output activations

    def get_edge_activations(self, l_i_j):
        """
        Returns a Tensor of Shape [n_samples] containing all activation values of Edge E_{l,i,j}
        :param l_i_j: Edge index
        :return: All Activation values of Edge with index l_i_j
        """
        l, i, j = l_i_j
        return self.edge_activations[l][:, i, j]

    def get_node_activations(self, l_i):
        """
        Returns a Tensor of Shape [n_samples] containing all activation values of Node N_{l,i}
        :param l_i: Node index
        :return: All Activation values of Node with index l_i
        """
        l, i = l_i
        return self.node_activations[l][:, i]

    def get_kan_layers(self):
        """
        Returns all Layers
        :return: Layers
        """
        return self.layers
    def get_layersizes(self):
        """
        Returns List of integers corresponding to the widths of each layer
        :return: list of layer widths
        """
        return self.layers_hidden
    def plot_curve(self, l_i_j_index):
        """
        Forwards the plot_curve call to the right layer (required by KanPlotter)
        :param l_i_j_index: Edge index
        :return:
        """
        l,i,j = l_i_j_index
        return self.layers[l].plot_curve(input_index=i, output_index=j)


class AttentionWithFastKANTransform(nn.Module):

    def __init__(
            self,
            q_dim: int,
            k_dim: int,
            v_dim: int,
            head_dim: int,
            num_heads: int,
            gating: bool = True,
    ):
        super(AttentionWithFastKANTransform, self).__init__()

        self.num_heads = num_heads
        total_dim = head_dim * self.num_heads
        self.gating = gating
        self.linear_q = FastKANLayer(q_dim, total_dim)
        self.linear_k = FastKANLayer(k_dim, total_dim)
        self.linear_v = FastKANLayer(v_dim, total_dim)
        self.linear_o = FastKANLayer(total_dim, q_dim)
        self.linear_g = None
        if self.gating:
            self.linear_g = FastKANLayer(q_dim, total_dim)
        # precompute the 1/sqrt(head_dim)
        self.norm = head_dim ** -0.5

    def forward(
            self,
            q: torch.Tensor,
            k: torch.Tensor,
            v: torch.Tensor,
            bias: torch.Tensor = None,  # additive attention bias
    ) -> torch.Tensor:

        wq = self.linear_q(q).view(*q.shape[:-1], 1, self.num_heads, -1) * self.norm  # *q1hc
        wk = self.linear_k(k).view(*k.shape[:-2], 1, k.shape[-2], self.num_heads, -1)  # *1khc
        att = (wq * wk).sum(-1).softmax(-2)  # *qkh
        del wq, wk
        if bias is not None:
            att = att + bias[..., None]

        wv = self.linear_v(v).view(*v.shape[:-2], 1, v.shape[-2], self.num_heads, -1)  # *1khc
        o = (att[..., None] * wv).sum(-3)  # *qhc
        del att, wv

        o = o.view(*o.shape[:-2], -1)  # *q(hc)

        if self.linear_g is not None:
            # gating, use raw query input
            g = self.linear_g(q)
            o = torch.sigmoid(g) * o

        # merge heads
        o = self.linear_o(o)
        return o


"""
File: csv_dataloader.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Dataloader and Dataset-Class for a comma-seperated CSV file or dictionary
"""
import pathlib
from pathlib import Path
import random
import pandas
import torch
import math
from torch.utils.data import Dataset


class CSVDataset(torch.utils.data.Dataset):
    def __init__(self,csv_file,input_vars,output_vars, categorical_vars, train_test="70:30", shuffle=True):
        """
        :param csv_file: file path to CSV file OR already loaded pandas dataframe OR dictionary
        :param input_vars: names of input variables (list)
        :param output_vars: names of output variables (list)
        :param categorical_vars: list of categorical variables that need to be converted to numeric values
        :param train_test: share of train and test data given as string 'Train:Test' i.e. '70:30'
        :param shuffle: flag to shuffle randomly before splitting the dataset
        """
        super().__init__()
        assert set(input_vars).isdisjoint(set(output_vars)), "Input and Output Variables cannot be overlapping sets!"
        self.input_varnames = input_vars
        self.output_varnames = output_vars
        self.categorical_vars = categorical_vars
        self.returnsplit = "train"  # The split self.__getitem__() should grab samples from.
        self.data = dict()
        print(
            f"Loading Dataset with input_vars={input_vars} and output_vars={output_vars} using categorical_vars={categorical_vars}.")
        # if csv_file is an actual file or a path
        if type(csv_file) == Path or type(csv_file) == pathlib.PosixPath:
            # load the file's contents using pandas
            print(f"Loading from Path {csv_file}")
            dataset_pandas = pandas.read_csv(csv_file)
            # restrict loaded dataset to input and output vars only
            dataset_pandas = dataset_pandas[input_vars + output_vars]
            # drop NA values
            dataset_pandas = dataset_pandas.dropna(axis=0, how="any")
        elif type(csv_file) == pandas.DataFrame:  # if csv_file is already a pandas dataset
            print("Using pre-loaded CSV Dataset passed as parameter")
            # just copy it
            dataset_pandas = csv_file
        elif type(csv_file) == dict:  # if csv_file is a dict
            # copy dict's contents to corresponding places in self.data
            self.data["train_input"] = csv_file["train_input"]
            self.data["train_label"] = csv_file["train_label"]
            self.data["test_input"] = csv_file["test_input"]
            self.data["test_label"] = csv_file["test_label"]
            return  # and return (we are done here in __init__()) !!!
        else:
            raise NotImplementedError(f"Dataset type {type(csv_file)} is not supported")
        # now dataset_pandas contains a pandas dataset that we need to convert to dictionary
        # create placeholder for train_input (for now everything ends up in train split)
        self.data["train_input"] = torch.zeros(size=(len(dataset_pandas), len(input_vars)), dtype=torch.float32)
        if len(output_vars)==1 and output_vars[0] in categorical_vars:  # Classification
            print("Amounts of samples per class: ", dataset_pandas[output_vars[0]].value_counts())
            self.tasktype = "classification"
            self.classes = list(dataset_pandas.groupby(output_vars[0])[output_vars[0]].unique())
            self.num_classes = len(self.classes)
            # create placeholder for One-Hot encoded classlabels
            self.data["train_label"] = torch.zeros(size=(len(dataset_pandas), self.num_classes), dtype=torch.float32)
            # output varnames are multiplied by num_classes because of One-Hot encoding
            self.output_varnames = self.num_classes * self.output_varnames
        else:  # Regression
            # create placeholder for output vars
            self.data["train_label"] = torch.zeros(size=(len(dataset_pandas), len(output_vars)), dtype=torch.float32)
            self.tasktype = "regression"

        # iterate over input_vars
        for i, col in enumerate(input_vars):
            if col in categorical_vars:  # if variable is categorical
                # encode categorical features as 0,1,2,...
                self.data["train_input"][:, i] = torch.from_numpy(
                    dataset_pandas[col].astype("category").cat.codes.values.copy())
                print("Converted categorical variables ", col, dataset_pandas[col].unique(), " to ",
                      torch.unique(self.data["train_input"][:, i]), " for col ", col)
            else:  # feature is numeric, so just copy the column into dict
                self.data["train_input"][:, i] = torch.from_numpy(dataset_pandas[col].values)
        # iterate over output_vars
        for i, col in enumerate(output_vars):
            if col in output_vars:
                if self.tasktype == "classification":
                    # encode categorical classnames as numbers 0,1,2,...
                    class_codes = torch.from_numpy(
                        dataset_pandas[col].astype("category").cat.codes.values.copy()).to(dtype=torch.int64)
                    # save the mapping (i.e. which class names ended up as which number)
                    class_classid_mapping = torch.unique(torch.stack([torch.from_numpy(dataset_pandas[col].to_numpy()), class_codes.squeeze()], dim=1), dim=0)
                    print("Converted categorical variables ", dataset_pandas[col].unique(), " to ",
                          torch.unique(class_codes), " for col ", col, "mapping", class_classid_mapping)
                    # convert encoded classnames into One-Hot encoded Tensors
                    self.data["train_label"][:, :] = torch.nn.functional.one_hot(class_codes, self.num_classes).float()
                else:
                    # if regression, just copy the target values into the right column
                    self.data["train_label"][:, i] = torch.from_numpy(dataset_pandas[col].values).float()

        # make sure everything is float
        self.data["train_input"] = self.data["train_input"].float()
        # calculate fraction of train and test
        train_share = int(train_test.split(":")[0])
        test_share = int(train_test.split(":")[1])
        train_fraction = train_share / (test_share + train_share)

        # Split into train and test
        indices = list(range(self.data["train_label"].size()[0]))  # all indices
        if shuffle:  # maybe randomly shuffle
            random.shuffle(indices)
        # draw test-samples out of train-split (until now everything was inside train-split)
        self.data["test_input"] = self.data["train_input"][indices[math.ceil(train_fraction * len(indices)):]]
        self.data["test_label"] = self.data["train_label"][indices[math.ceil(train_fraction * len(indices)):]]
        self.data["train_input"] = self.data["train_input"][indices[:math.ceil(train_fraction * len(indices))]]
        self.data["train_label"] = self.data["train_label"][indices[:math.ceil(train_fraction * len(indices))]]

    def to(self, device):
        """
        Move dataset to device
        :param device: device
        """
        if type(device) == str:
            device = torch.device(device)
        self.data["test_input"] = self.data["test_input"].to(device)
        self.data["test_label"] = self.data["test_label"].to(device)
        self.data["train_input"] = self.data["train_input"].to(device)
        self.data["train_label"] = self.data["train_label"].to(device)
    def get_train_test_size(self):
        """
        Get train and test size (number of samples)
        :return: Tuple (train_size, test_size)
        """
        return self.data["train_input"].size()[0], self.data["test_input"].size()[0]

    def __or__(self, other):
        """
        Unite two CSVDatasets with each other
        :param other: the other CSVDataset to unite with
        :return: joined dataset
        """
        assert type(other) == CSVDataset, "Expected other dataset to unite with to also be of class CSVDataset!"
        assert other.input_varnames == self.input_varnames, "Can only unite datasets with same input_varnames"
        assert other.output_varnames == self.output_varnames, "Can only unite datasets with same output_varnames"
        assert other.categorical_vars == self.categorical_vars, "Can only unite datasets with same categorical_vars"
        # calculate new number of samples in train and test split (add them together)
        new_train_length = self.get_train_test_size()[0] + other.get_train_test_size()[0]
        new_test_length = self.get_train_test_size()[1] + other.get_train_test_size()[1]

        print(f"Joining Dataset of size {self.get_train_test_size()} with Dataset of size {other.get_train_test_size()}...")
        print("Be careful that all categorical values have the same corresponding codes!")
        # create placeholders for the joined dataset's samples
        new_train_input =torch.zeros(size=(new_train_length, len(self.input_varnames)), dtype=torch.float32)
        new_train_label=torch.zeros(size=(new_train_length, len(self.output_varnames)), dtype=torch.float32)
        new_test_input =torch.zeros(size=(new_test_length, len(self.input_varnames)), dtype=torch.float32)
        new_test_label=torch.zeros(size=(new_test_length, len(self.output_varnames)), dtype=torch.float32)

        # copy everything from self to new dataset
        new_train_input[0:self.data["train_input"].size()[0], :] = self.data["train_input"]
        new_train_label[0:self.data["train_label"].size()[0], :] = self.data["train_label"]
        new_test_input[0:self.data["test_input"].size()[0], :] = self.data["test_input"]
        new_test_label[0:self.data["test_label"].size()[0], :] = self.data["test_label"]

        # add every sample from 'other' to new dataset
        new_train_input[self.data["train_input"].size()[0]:, :] = other.data["train_input"]
        new_train_label[self.data["train_label"].size()[0]:, :] = other.data["train_label"]
        new_test_input[self.data["test_input"].size()[0]:, :] = other.data["test_input"]
        new_test_label[self.data["test_label"].size()[0]:, :] = other.data["test_label"]

        # build dictionary
        joined_dataset = dict()
        joined_dataset["train_input"] = new_train_input
        joined_dataset["train_label"] = new_train_label
        joined_dataset["test_input"] = new_test_input
        joined_dataset["test_label"] = new_test_label
        joined_dataset = CSVDataset(joined_dataset, self.input_varnames, self.output_varnames, self.categorical_vars)
        joined_dataset.returnsplit = self.returnsplit
        if hasattr(self.num_classes):
            joined_dataset.num_classes = self.num_classes
        joined_dataset.tasktype = self.tasktype
        print(f"Joined Dataset has size {joined_dataset.get_train_test_size()} (Train, Test)")
        return joined_dataset


    def __len__(self):
        """
        Get length of dataset
        :return: length of dataset
        """
        # return length for selected split
        if self.returnsplit == "train":
            return self.data["train_input"].size()[0]
        elif self.returnsplit == "test":
            return self.data["test_input"].size()[0]
    def __getitem__(self, index):
        """
        Get sample at index 'index'
        :param index: index to get sample from
        :return: sample as a tuple (x,y)
        """
        # return sample from selected split
        if self.returnsplit == "train":
            return self.data["train_input"][index], self.data["train_label"][index]
        if self.returnsplit == "test":
            return self.data["test_input"][index], self.data["test_label"][index]
    def set_returnsplit(self, return_split):
        """
        Update the dataset's split to return when used in a dataloader
        :param return_split: new returnsplit. Either 'train' or 'test'
        """
        self.returnsplit = return_split
    def normalize(self):
        """Scale and shift each feature seperately so that all values of that feature fall in range  [-2, 2]"""
        # iterate over all features
        for i, feature_name in enumerate(self.input_varnames):
            # collect mins and maxs across both train and test split, if each not empty
            mins = []
            maxs = []
            # Train and Test set might be empty
            if self.get_train_test_size()[1] > 0:
                mins.append(torch.amin(self.data["test_input"][:,i]))
                maxs.append(torch.amax(self.data["test_input"][:, i]))
            if self.get_train_test_size()[0] > 0:
                mins.append(torch.amin(self.data["train_input"][:,i]))
                maxs.append(torch.amax(self.data["train_input"][:, i]))
            min_ = torch.min(torch.Tensor(mins))
            max_ = torch.max(torch.Tensor(maxs))
            # scale so that current feature covers interval [-2, 2] completely (i.e. shift so that new min is zero,
            # then scale to [0, 1] and then re-scale to [-2, 2]
            if self.get_train_test_size()[1] > 0:
                self.data["test_input"][:, i] = (self.data["test_input"][:, i] - min_) / (max_ - min_) * 4 - 2
            if self.get_train_test_size()[0] > 0:
                self.data["train_input"][:,i] = (self.data["train_input"][:,i] - min_) / (max_ - min_) * 4 - 2
    def normalize_for_pykan(self):
        """
        Normalize dataset for pyKAN, which expects dataset to have std of 1 and mean of zero.
        """
        mean_train = torch.mean(self.data["train_input"], dim=0)
        std_train = torch.std(self.data["train_input"], dim=0)

        self.data["train_input"] = (self.data["train_input"] - mean_train) / std_train
        if self.get_train_test_size()[1] > 0:
            self.data["test_input"] = (self.data["test_input"] - mean_train) / std_train


"""
File: latex_table_creator.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Automatically create LaTeX Tables from results.json logfile
"""
import numpy as np
import pandas as pd
import json
from pathlib import Path

def get_dataset_name(json_data):
    """
    Returns dataset name
    :param json_data: JSON entry for current experiment
    :return: dataset name
    """
    res = json_data["dataset_name"]
    # convert internal names (i.e. ff_mult) to proper names
    replacement_dict = {"ff_mult": "$z_1*z_2$", "ff_sin": "$\sin (z)$", "ff_square": "$z^2$", "ff_squaresquare": "$(z_1^2 + z_2^2)^2$",
                        "ph_holo_r_100k": "Holography", "ph_circuit_r_100k": "Circuit", "ph_holo_c_100k": "Holography", "ph_circuit_c_100k": "Circuit"}
    return replacement_dict[res] if res in replacement_dict else res
def get_model_name(json_data):
    """
    Returns model name
    :param json_data: JSON entry for current experiment
    :return: model name
    """
    res = json_data["model_name"]
    # convert internal model names to more understandable names
    replacement_dict = {"FastKAN": "FastKAN", "CVKANWrapper": "$\C$KAN", "PyKANWrapper": "KAN"}
    return replacement_dict[res] if res in replacement_dict else res
def get_model_size(json_data):
    """
    Returns size of model (layer_0 x layer_1 x ... x layer_N) as a string
    :param json_data: JSON entry for current experiment
    :return: size of model (layer_0 x layer_1 x ... x layer_N)
    """
    return "$" + " {\\times} ".join([str(l) for l in json_data["functions"]]) + "$"
def get_model_params(json_data):
    """
    Returns number of trainable parameters
    :param json_data: JSON entry for current experiment
    :return: number of trainable parameters
    """
    return str(json_data["num_trainable_params"])
def get_model_normalization(json_data):
    """
    Returns normalization scheme used during current experiment
    :param json_data: JSON entry for current experiment
    :return: normalization scheme used
    """
    res = json_data["use_norm"]
    # assume that all layers used the same normalization scheme; thus just take the normalization scheme of first layer
    if type(res) == list:
        res = res[0]
    # replace internal names with something more understandable
    replacement_dict = {True: "BN", False : "no", "batchnorm": "$\\text{BN}_\\C$", "batchnormvar": "$\\text{BN}_\\mathbb{V}$", "batchnormnaiv": "$\\BN_{\\R^2}$", None: "no", "grid_update": "grid\_update"}
    return replacement_dict[res] if res in replacement_dict else res
def get_model_num_grids(json_data):
    """
    Returns number of gridpoints used
    :param json_data: JSON entry for current experiment
    :return: number of gridpoints used
    """
    return json_data["num_grids"]
def get_model_zsilu_type(json_data):
    """
    Returns type of complex-SiLU used during current experiment
    :param json_data: JSON entry for current experiment
    :return: type of complex-SiLU used
    """
    res = json_data["zsilu_type"]
    # replace internal names with something more understandable
    replacement_dict = {"complex_weight": "c", "real_weights": "r", None: "-"}
    return replacement_dict[res] if res in replacement_dict else res
def get_test_acc(json_data):
    """
    Returns accuracy on test-set averaged over all folds including standard deviation as a string
    :param json_data: JSON entry for current experiment
    :return: string for test accuracy in the form of '0.98 +- 0.02'
    """
    return ("$" +
            "{:.3f}".format(json_data["test_losses"]["mean"]["accuracy"]) +
             " $\\scriptsize{$\\pm " +
             "{:.3f}".format(json_data["test_losses"]["std"]["accuracy"]) +
            "$}")
def get_test_ce(json_data):
    """
    Returns Cross-Entropy Loss on test-set averaged over all folds including standard deviation as a string
    :param json_data: JSON entry for current experiment
    :return: string for test CE-Loss in the form of '0.12 +- 0.02'
    """
    return ("$" +
            "{:.3f}".format(json_data["test_losses"]["mean"]["cross_entropy"]) +
             " $\\scriptsize{$\\pm " +
             "{:.3f}".format(json_data["test_losses"]["std"]["cross_entropy"]) +
            "$}")
def get_mae(json_data, split="train"):
    """
    Returns MAE on specified split averaged over all folds including standard deviation as a string
    :param json_data: JSON entry for current experiment
    :param split: split to calculate MAE on
    :return: MAE in the form of '0.12 +- 0.02' for selected split
    """
    return ("$" +
            "{:.3f}".format(json_data[split+"_losses"]["mean"]["mae"]) +
             " $\\scriptsize{$\\pm " +
             "{:.3f}".format(json_data[split+"_losses"]["std"]["mae"]) +
            "$}")
def get_mse(json_data, split="test"):
    """
    Returns MSE on specified split averaged over all folds including standard deviation as a string
    :param json_data: JSON entry for current experiment
    :param split: split to calculate MSE on
    :return: MSE in the form of '0.12 +- 0.02' for selected split
    """
    return ("$" +
            "{:.3f}".format(json_data[split+"_losses"]["mean"]["mse"]) +
             " $ \\scriptsize{$\\pm " +
             "{:.3f}".format(json_data[split+"_losses"]["std"]["mse"]) +
            "$}")

def create_table(results_json, column_titles_latex, column_values, filter_func=None, sort_by=None):
    """
    Creates a LaTeX table from given json and prints it to shell
    :param results_json: JSON from whole results.json
    :param column_titles_latex: list of title names of the columns in the LaTeX table
    :param column_values: List of functions that extract the wanted value from each list entry inside the results_json
    :param filter_func: Filter function to apply for specifying only certain experiments / results
    :param sort_by: sort by which columns
    """
    assert len(results_json) > 0
    # if no filter function specified, make filter always return True
    if filter_func is None:
        filter_func = lambda x: True  # filter = always True
    # build table
    rows = []
    for entry in results_json:
        # if current entry (json) matches filter_func
        if filter_func(entry):
            # extract values from json by applying column_values() lambda
            row = column_values(entry)
            assert len(column_titles_latex) == len(row)
            rows.append(row)
    rows = np.array(rows)
    # convert this table to pandas dataframe
    data_dict = dict()
    for i in range(len(column_titles_latex)):
        data_dict[column_titles_latex[i]] = rows[:, i]
    df = pd.DataFrame(data_dict)

    # now create a LaTeX table based on the dataframe
    finished_latex = ""
    preamble = "\\begin{tabular}{" + len(df.columns) * "|c" + "|" + "}\n\\hline\n"
    finished_latex += preamble
    # Create top row (column names)
    top_row = ""
    for col_name in column_titles_latex:
        top_row += col_name + " & "
    top_row = top_row[:-2] + "\\\\\\hline\n"
    finished_latex += top_row
    # potentially sort values by (multiple) column name(s)
    if sort_by is not None:
        df = df.sort_values(by=sort_by)
    table_np = df.to_numpy()
    # iterate through the rows of the table
    for row in range(table_np.shape[0]):
        # build LaTeX string for current row
        row_latex = ""
        for col in range(table_np.shape[1]):
            row_latex += str(table_np[row, col]) + " & "
        # remove last "&"
        row_latex = row_latex[:-2]
        # add linebreak
        row_latex += "\\\\"
        # add newline
        finished_latex += row_latex + "\n"
    # finish LaTeX table
    finished_latex += "\\hline\n"
    finished_latex += "\n\\end{tabular}\n"
    # print LaTeX table
    print(finished_latex)


def knot_dataset_table():
    """
    Creates a LaTeX table for the knot dataset
    """
    # column titles
    column_titles_latex = ["Model",
                           "Size",
                           "\# Params",
                           "Normalization",
                           #"\# Grids",
                           "zSiLU",
                           "Test Acc.",
                           "Test CE-Loss"]
    # corresponding (tuple of) functions to extract specific values from json entry
    column_values = lambda x: (get_model_name(x),
                               get_model_size(x),
                               get_model_params(x),
                               get_model_normalization(x),
                               #get_model_num_grids(x),
                               get_model_zsilu_type(x),
                               get_test_acc(x),
                               get_test_ce(x))
    # filter for knot dataset only (and CVKAN)
    filter_func = lambda x: x["dataset_name"].startswith("knot_") and x["model_name"] == "CVKANWrapper"
    # create LaTeX table
    create_table(results_json, column_titles_latex=column_titles_latex, column_values=column_values,
                 filter_func=filter_func, sort_by=["Model", "Size"])

def function_fitting_table():
    """
    Creates a LaTeX table for the fitting function datasets
    """
    # column titles
    column_titles_latex = [#"Dataset",
                           "Model",
                           "Size",
                           "\# Params",
                           #"\# Grids",
                           #"zSiLU",
                           "Test MSE",
                           "Test MAE",
                           #"Train MSE",
                           #"Train MAE",
                           ]
    # corresponding (tuple of) functions to extract specific values from json entry
    column_values = lambda x: (#get_dataset_name(x),
                               get_model_name(x),
                               get_model_size(x),
                               get_model_params(x),
                               #get_model_num_grids(x),
                               #get_model_zsilu_type(x),
                               get_mse(x, split="test"),
                               get_mae(x, split="test"),
                               #get_mse(x, split="train"),
                               #get_mae(x, split="train")
                               )
    # filter for physical circuit dataset with 100k samples
    filter_func = lambda x: x["dataset_name"].startswith("ph_circ") and x["dataset_name"].endswith("100k")
    # filter for general function fitting datasets (synthetic)
    #filter_func = lambda x: x["dataset_name"].startswith("ff_")
    # create LaTeX table
    create_table(results_json, column_titles_latex=column_titles_latex, column_values=column_values,
                 filter_func=filter_func, sort_by=["Model"])



if __name__ == "__main__":
    p = Path("/src/experiments/results.json")
    with open(p, "r", encoding="utf-8") as f:
        results_json = json.load(f)

    knot_dataset_table()
    #function_fitting_table()





"""
File: misc.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Miscellaneous short methods
"""
import torch


def get_num_parameters(model):
    """
    Calculate number of parameters of a model
    :param model: model
    :return: number of parameters in model
    """
    sum_ = 0
    for param in model.parameters():
        if param.requires_grad:  # only sum up values that require grad
            if param.dtype == torch.complex64:  # if type is complex-valued: multiply number of parameters by 2
                # complex parameter equivalent to 2 real-valued parameters
                sum_ += param.numel() * 2
            else:
                sum_ += param.numel()
    return sum_


def mean(l):
    """
    Calculate the mean of a list of numbers
    :param l: list
    :return: mean of list
    """
    return sum(l) / len(l)


"""
File: loss_functions.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Different loss-functions (MSE, MAE and cross entropy)
"""
import numpy as np
import torch


def mse_formula(pred, label):
    """
    Calculates MSE between prediction and label
    :param pred: prediction
    :param label: label
    :return: MSE value
    """
    pred = pred.squeeze()
    label = label.squeeze()
    assert pred.shape == label.shape, f"Expected Shape {label.shape}, got {pred.shape}"
    if type(pred) == np.ndarray:  # do calculation within numpy
        return np.mean((np.abs(pred-label))**2)
    if type(pred) == torch.Tensor:  # do calculation within torch
        return torch.mean((torch.abs(pred-label))**2)


def mae_formula(pred, label):
    """
    Calculates MAE between prediction and label
    :param pred: prediction
    :param label: label
    :return: MAE value
    """
    pred=pred.squeeze()
    label=label.squeeze()
    assert pred.shape == label.shape
    if type(pred) == np.ndarray:  # do calculation within numpy
        return np.mean(np.abs(pred-label))
    elif type(pred) == torch.Tensor:  # do calculation within torch
        return torch.mean(torch.abs(pred-label))


def cross_entropy(pred, label):
    """
    Calculates cross entropy between prediction and label
    :param pred: prediction
    :param label: label
    :return: cross entropy value
    """
    return torch.nn.functional.cross_entropy(pred, label)


class MSE(torch.nn.Module):
    """
    MSE as a torch module
    """
    def __init__(self):
        super().__init__()
    def forward(self, pred, label):
        return torch.mean((torch.abs(pred - label)) ** 2)


class MAE(torch.nn.Module):
    """
    MAE as a torch module
    """
    def __init__(self):
        super().__init__()
    def forward(self, pred, label):
        return torch.mean(torch.abs(pred - label))



"""
File: crossval_splitter.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Convert dataset with 100% train split to k datasets with non-overlapping test-splits of size (1/k)
using k-fold cross-validation
"""
import torch
from src.cvkan.utils.dataloading.csv_dataloader import CSVDataset

def split_crossval(dataset: CSVDataset, k=5):
    """

    :param dataset: CSVDataset dataset with 100% train split
    :param k: number of folds to do
    :return: list of k datasets, each element representing one fold.
    """
    assert dataset.get_train_test_size()[1] == 0, "For automatic crossvalidation splitting the test set should be empty (100% train)"
    dataset_size = dataset.get_train_test_size()[0]
    crossval_datasets = []
    # loop over all possible split points (0,1,2,...,k-1)
    for split in range(k):
        # calculate test start and end indices
        test_start_index = split * (dataset_size//k)
        test_end_index = (split + 1) * (dataset_size//k)
        # create new dataset dictionary
        data = dict()
        # copy test set from original dataset for current fold based on previously calculated range of indices
        data["test_input"] = dataset.data["train_input"][test_start_index:test_end_index]
        data["test_label"] = dataset.data["train_label"][test_start_index:test_end_index]
        # if test split is surrounded by train data
        if test_start_index > 0 and test_end_index < dataset_size:
            # copy data before and after the test split into the train split
            data["train_input"] = torch.cat((dataset.data["train_input"][0:test_start_index], dataset.data["train_input"][test_end_index:dataset_size, :]))
            data["train_label"] = torch.cat((dataset.data["train_label"][0:test_start_index], dataset.data["train_label"][test_end_index:dataset_size, :]))
        # if test split is the last split
        elif test_start_index > 0 and test_end_index == dataset_size:
            # copy eberything before test split to train split
            data["train_input"] = dataset.data["train_input"][0:test_start_index, :]
            data["train_label"] = dataset.data["train_label"][0:test_start_index, :]
        # if test split is the first split
        elif test_start_index == 0 and test_end_index != dataset_size:
            # copy everything after test split to train split
            data["train_input"] = dataset.data["train_input"][test_end_index:, :]
            data["train_label"] = dataset.data["train_label"][test_end_index:, :]
        # create a CSVDataset object out of the constructed dictionary, copying varnames
        ds = CSVDataset(data, input_vars=dataset.input_varnames, output_vars=dataset.output_varnames, categorical_vars=dataset.categorical_vars)
        # set attribute num_classes the same as in the original datasets, if exists
        if hasattr(dataset, "num_classes"):
            ds.num_classes = dataset.num_classes
        # append current fold's dataset to the list of datasets
        crossval_datasets.append(ds)
    # return list of k datasets
    return crossval_datasets



"""
File: run_crossval.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Running k-fold cross-validation training runs and logging the results with additional meta-data
             in a json file
"""
import copy
import json
import os
from datetime import datetime
from pathlib import Path
import numpy as np
import torch

from ..train.train_loop import train_kans
from ..utils.dataloading.crossval_splitter import split_crossval
from ..utils.dataloading.csv_dataloader import CSVDataset
from ..utils.misc import get_num_parameters


def run_crossval(model, dataset_full_train: CSVDataset, dataset_name, loss_fn_backprop, loss_fns, batch_size, device=torch.device("cuda"), logging_interval=100, add_softmax_lastlayer=False, epochs=500, convert_model_output_to_real=True, k=5):
    """
    Runs cross-validaiton on a given dataset
    :param model: Model to train
    :param dataset_full_train: CSVDataset object that only contains training data and zero test data for splitting into k folds
    :param dataset_name: Name of the dataset (for logging purposes only)
    :param loss_fn_backprop: Loss function to use for backpropagation
    :param loss_fns: dictionary of additional loss functions (also including loss_fn_backprop) to evaluate and save
    :param batch_size: batch size to use for training. Should be -1 for pyKAN models!
    :param device: device to train on
    :param logging_interval: how often to display logs in console
    :param add_softmax_lastlayer: flag whether softmax should be added after the last layer
    :param epochs: epochs to train for
    :param convert_model_output_to_real: should be True only if model produces complex-valued output but
    we need real-valued output. Essentially does model(x).real instead of only model(x)
    :param k: number of folds to do cross-validation on
    :return: None
    """
    # generate k-fold cv (usually k=5)
    datasets = split_crossval(dataset_full_train, k=k)  # returns list of datasets (with different crossval splits each)
    num_folds = len(datasets)
    # build dictionary to store the results into
    results = dict()
    results["train_losses"] = dict()
    results["test_losses"] = dict()
    # mean and std across the k folds
    results["train_losses"]["mean"] = dict()
    results["train_losses"]["std"] = dict()
    results["test_losses"]["mean"] = dict()
    results["test_losses"]["std"] = dict()
    results["model_name"] = model.__class__.__name__
    results["dataset_size"] = dataset_full_train.get_train_test_size()
    # make sure loss_fn_backprop is a class and not just a lambda or method.
    results["loss_fn_backprop"] = loss_fn_backprop.__class__.__name__
    results["batch_size"] = batch_size
    results["epochs"] = epochs
    results["complex"] = convert_model_output_to_real
    results["add_softmax_lastlayer"] = add_softmax_lastlayer
    results["layers"] = model.layers_hidden
    results["use_norm"] = model.use_norm
    results["num_grids"] = model.num_grids
    results["rho"] = model.rho if hasattr(model, "rho") else None
    results["dataset_name"] = dataset_name
    results["kfolds"] = k
    results["num_trainable_params"] = get_num_parameters(model)
    results["zsilu_type"] = model.csilu_type if hasattr(model, "csilu_type") else None
    results["start_timestamp"] = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    # iterate over the k folds
    for i, d in enumerate(datasets):
        results["train_losses"][i] = dict()
        results["test_losses"][i] = dict()
        my_model = copy.deepcopy(model)  # create a copy of the model
        # run training on current fold
        train_loss, test_loss = train_kans(my_model, d, loss_fn_backprop=loss_fn_backprop, loss_fns=loss_fns,
                                           device=device, batch_size=batch_size, logging_interval=logging_interval,
                                           add_softmax_lastlayer=add_softmax_lastlayer, epochs=epochs,
                                           last_layer_output_real=convert_model_output_to_real)
        # and store results for each loss_fn given
        for k in loss_fns.keys():
            results["train_losses"][i][k] = train_loss[k].item()
            results["test_losses"][i][k] = test_loss[k].item()

    # calculate mean and std for all loss_fn's given
    for k in loss_fns.keys():
        train_losses_per_function = np.array([results["train_losses"][i][k] for i in range(num_folds)])
        test_losses_per_function = np.array([results["test_losses"][i][k] for i in range(num_folds)])
        results["train_losses"]["mean"][k] = np.mean(train_losses_per_function)
        results["train_losses"]["std"][k] = np.std(train_losses_per_function)
        results["test_losses"]["mean"][k] = np.mean(test_losses_per_function)
        results["test_losses"]["std"][k] = np.std(test_losses_per_function)
    results["finish_timestamp"] = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    print(results)
    results_file = Path(os.path.abspath(__file__)).parent / "results.json"
    if results_file.exists():
        with open(results_file, "r", encoding="utf-8") as f:
            all_results = json.load(f)
    else:
        all_results = []
    all_results.append(results)
    with open(results_file, "w+", encoding="utf-8") as f:
        json.dump(all_results, f)



"""
File: eval_model.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Methods to evaluate a given model on arbitrary loss functions as well as for plotting a confusion matrix
for Predictions / GT.
"""
import matplotlib
import numpy as np
import sklearn
import torch
from matplotlib import pyplot as plt
from sklearn.metrics import confusion_matrix

def eval_model(model, loss_fns, *, train_data, test_data, train_label, test_label, add_softmax_lastlayer):
    """
    Evaluates a given model on given train and test data as well as ground-truth labels using a dictionary of loss functions
    :param model: model to evaluate
    :param loss_fns: dictionary of loss functions to evaluate the model on
    :param train_data: train input samples (no batching)
    :param test_data: test input samples (no batching)
    :param train_label: train ground truth labels (no batching)
    :param test_label: test ground truth labels (no batching)
    :param add_softmax_lastlayer: whether softmax should be applied after the last layer before evaluation of loss functions
    (CE loss named 'cross_entropy' will ignore this value and always receive raw logits)
    :return: Tuple of two dictionaries (train_losses, test_losses) that each have the names of the loss functions
    evaluated as keys and the resulting value of each loss function as value.
    """
    # dictionaries to store train and test losses for all loss functions to evaluate
    train_losses = dict()
    test_losses = dict()
    # create predictions for train and test input data
    train_predictions = model(train_data)
    test_predictions = model(test_data)
    # iterate over all loss functions
    for loss_fn_name, loss_fn in loss_fns.items():
        current_train_label = train_label
        current_test_label = test_label
        current_train_preds = train_predictions
        current_test_preds = test_predictions
        # if loss function is CE, accuracy, F1 or precision and model output is complex
        if loss_fn_name in ["cross_entropy", "accuracy", "f1", "precision"] and train_predictions.dtype == torch.complex64:
            # only use real parts
            current_train_preds = train_predictions.real
            current_test_preds = test_predictions.real
            # if softmax should be applied (and loss function is not CE; CE requires raw logits)
        if add_softmax_lastlayer and loss_fn_name != "cross_entropy":
            # apply softmax
            current_train_preds = torch.nn.functional.softmax(current_train_preds, dim=1)
            current_test_preds = torch.nn.functional.softmax(current_test_preds, dim=1)
        # some loss functions require argmax and not softmax values
        if loss_fn_name in ["accuracy", "f1", "precision"]:
            current_train_label = torch.argmax(current_train_label, dim=1)
            current_test_label = torch.argmax(current_test_label, dim=1)
            current_train_preds = torch.argmax(current_train_preds, dim=1)
            current_test_preds = torch.argmax(current_test_preds, dim=1)
        # insert train and test losses into the dictionaries with key = name of loss function
        train_losses[loss_fn_name] = loss_fn(current_train_preds, current_train_label)
        test_losses[loss_fn_name] = loss_fn(current_test_preds, current_test_label)
    return train_losses, test_losses

def plot_confusion_matrix(pred, gt, labelmapping=None):
    """
    Plot a confusion matrix based on Predictions and GT values
    :param pred: Predictions
    :param gt: Labels
    :param labelmapping: List of class names in order of class ids for axis descriptions
    """
    # set plot parameters & create figure
    font = {'family': 'normal',
            'weight': 'bold',
            'size': 11}
    matplotlib.rc('font', **font)
    plt.figure(figsize=(16, 16))
    # potentially complex predictions are converted to real part only
    if pred.dtype == torch.complex64:
        pred = pred.real
    # if predictions are softmax values, convert to class ids (argmax)
    if len(pred.shape) > 1:
        pred = pred.argmax(axis=1)
    # if labels are One-Hot encoded values, convert to class ids (argmax)
    if len(gt.shape) > 1:
        gt = gt.argmax(axis=1)
    pred = pred.detach().cpu().numpy()
    gt = gt.detach().cpu().numpy()
    # build confusion matrix
    cm = confusion_matrix(y_true=gt, y_pred=pred, normalize="true",labels=[i for i in range(len(labelmapping))] if labelmapping is not None else None)
    # round to 2 decimal places
    cm = np.round(cm, decimals=2)
    # display confusion matrix
    cmd = sklearn.metrics.ConfusionMatrixDisplay(cm, display_labels=labelmapping)
    cmd.plot()
    plt.xlabel('Predicted label', fontsize=20)
    plt.ylabel('True label', fontsize=20)
    #plt.savefig('images/cvkan_knot_confusionmatrix.svg', transparent=True, bbox_inches='tight', pad_inches=0.5)
    plt.show()


from .models.CVKAN import CVKAN
from .models.wrapper.WrapperTemplate import WrapperTemplate
from .models.wrapper.CVKANWrapper import CVKANWrapper
from .utils.explain_kan import KANExplainer
from .utils.plotting.plot_kan import KANPlot
from .train.train_loop import train_kans

__all__ = [CVKAN, CVKANWrapper, KANExplainer, KANPlot, train_kans]


"""
File: CV_LayerNorm.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Three different complex-valued BatchNorm approaches and one complex-valued LayerNorm
"""
import torch.nn as nn
import torch


class Complex_LayerNorm(nn.Module):  # this is an alternate implementation of Layernorm where everything is wrapped in one module, the split one is preferable for explainability methods

    def __init__(self, device='cuda'):

        super().__init__()

    def forward(self, input):

        ev = torch.unsqueeze(torch.mean(input, dim=1), dim=1)

        cov_shape = [3, input.shape[0], 1]
        cov_m = torch.zeros(cov_shape, device=input.device)
        cov_m[0] = torch.unsqueeze(torch.var(input.real, dim=1), dim=1) + 1e-5  # 1e-5 if variance 0
        cov_m[1] = torch.unsqueeze(torch.var(input.imag, dim=1), dim=1) + 1e-5  # 1e-5 if variance 0

        input = input - ev

        cov_m[2] = torch.unsqueeze(torch.mean(input.real * input.imag, dim=1), dim=1)  # cov(real, imag)

        cov_m = self.inv_sqrt_2x2(cov_m)

        input = self.mult_2x2_same_dim(input, cov_m)  # decorrelate input

        return input

    def inv_sqrt_2x2(self, input):
        input = torch.unsqueeze(input, dim=0)
        s = torch.sqrt(input[:, 0] * input[:, 1] - input[:, 2] ** 2)
        t = torch.sqrt(input[:, 0] + input[:, 1] + 2 * s)
        return 1 / (t * s) * torch.cat([input[:, 1] + s, input[:, 0] + s, -input[:, 2]], dim=0)

    def sqrt_2x2(self, input):
        input = torch.unsqueeze(input, dim=0)
        s = torch.sqrt(input[:, 0] * input[:, 1] - input[:, 2] ** 2)
        t = torch.sqrt(input[:, 0] + input[:, 1] + 2 * s)
        return 1 / t * torch.cat([input[:, 0] + s, input[:, 1] + s, input[:, 2]], dim=0)

    def mult_2x2_same_dim(self, input, mult):
        return mult[0] * input.real + mult[2] * input.imag + 1j * (mult[2] * input.real + mult[1] * input.imag)


class Complex_BatchNorm(nn.Module):
    def __init__(self, num_channel, affine=True, momentum=0.1, device='cuda'):
        super().__init__()
        self.register_buffer('running_mean', tensor=torch.zeros([1, num_channel], dtype=torch.complex64, requires_grad=False))
        self.register_buffer('running_cov', tensor=torch.Tensor([1, 1, 0]).repeat([1, num_channel]))
        self.device = device
        self.zeros_cpu = torch.zeros(5, device=device)
        self.zeros_gpu = torch.zeros(5, device=self.device)
        self.momentum = momentum
        self.first_run = True
        self.affine = affine
        if self.affine:
            self.weight = nn.Parameter(torch.cat([torch.ones([2, num_channel], dtype=torch.float32, device=self.device), torch.zeros([1, num_channel], dtype=torch.float32, device=self.device)], dim=0))
            self.bias = nn.Parameter(torch.zeros(num_channel, dtype=torch.complex64, device=self.device))

    def forward(self, input):
        if self.training:

            ev = torch.unsqueeze(torch.mean(input, dim=0), dim=0)

            with torch.no_grad():
                if not self.first_run:
                    self.running_mean = (ev * (1 - self.momentum) + self.running_mean * self.momentum).detach()
                else:
                    self.running_mean = ev.detach()

            cov_m = torch.zeros([3, input.shape[1]], device=input.device)
            cov_m[0] = torch.var(input.real, dim=0) + 1e-5
            cov_m[1] = torch.var(input.imag, dim=0) + 1e-5  # 1e-5 if variance 0

            input = input - ev
            cov_m[2] = torch.mean(input.real * input.imag, dim=0)

            cov_m = self.inv_sqrt_2x2(cov_m)
            with torch.no_grad():
                if not self.first_run:
                    self.running_cov = (cov_m * (1 - self.momentum) + self.running_cov * self.momentum).detach()  # note: running_cov is already sqrt and inv
                else:
                    self.running_cov = cov_m.detach()
                    self.first_run = False

        else:
            cov_m = self.running_cov.detach()  # cov is already saved as inv_sqrt
            ev = self.running_mean.detach()

            input = input - ev

        input = self.mult_2x2(input, cov_m)  # decorrelate input

        if self.affine:
            # weight_calc = torch.zeros_like(self.weight)
            weight_var_real = self.weight[0]**2  # ensures varR>0
            weight_var_imag = self.weight[1]**2  # ensures varI>0
            weight_cov_ri = torch.sqrt(weight_var_real * weight_var_imag) * (torch.sigmoid(self.weight[2]) * 2 - 1)  # ensures covRI**2 < varR * varI
            weight_calc = torch.cat([weight_var_real.unsqueeze(0), weight_var_imag.unsqueeze(0), weight_cov_ri.unsqueeze(0)], dim=0)

            input = self.mult_2x2(input, weight_calc)
            input = input + self.bias

        return input

    def mult_2x2(self, input, mult):
        mult = torch.unsqueeze(mult, dim=0)
        return mult[:, 0] * input.real + mult[:, 2] * input.imag + 1j * (mult[:, 2] * input.real + mult[:, 1] * input.imag)

    def inv_sqrt_2x2(self, input):
        input = torch.unsqueeze(input, dim=0)
        s = torch.sqrt(input[:, 0] * input[:, 1] - input[:, 2] ** 2)
        t = torch.sqrt(input[:, 0] + input[:, 1] + 2 * s)
        return 1 / (t * s) * torch.cat([input[:, 1] + s, input[:, 0] + s, -input[:, 2]], dim=0)

    def sqrt_2x2(self, input):
        input = torch.unsqueeze(input, dim=0)
        s = torch.sqrt(input[:, 0] * input[:, 1] - input[:, 2] ** 2)
        t = torch.sqrt(input[:, 0] + input[:, 1] + 2 * s)
        return 1 / t * torch.cat([input[:, 0] + s, input[:, 1] + s, input[:, 2]], dim=0)


class Complex_BatchNorm_naiv(nn.Module):

    def __init__(self, num_channel, affine=True, eps=1e-05, momentum=0.1, device='cuda', dtype=torch.float32):
        super().__init__()
        self.num_channel = num_channel
        self.eps = eps
        self.momentum = momentum
        self.device = device
        self.dtype = dtype
        self.affine = affine

        self.real_BatchNorm = nn.BatchNorm1d(self.num_channel, eps=self.eps, momentum=self.momentum, affine=affine, track_running_stats=True, device=self.device, dtype=self.dtype)
        self.imag_BatchNorm = nn.BatchNorm1d(self.num_channel, eps=self.eps, momentum=self.momentum, affine=affine, track_running_stats=True, device=self.device, dtype=self.dtype)

    def forward(self, input):
        return torch.complex(self.real_BatchNorm(input.real), self.imag_BatchNorm(input.imag))


class Complex_BatchNorm_var(nn.Module):
    def __init__(self, num_channel, affine=True, momentum=0.1, device='cuda'):
        super().__init__()
        self.register_buffer('running_mean', tensor=torch.zeros([1, num_channel], dtype=torch.complex64, requires_grad=False))
        self.register_buffer('running_var', tensor=torch.ones([1, num_channel], dtype=torch.complex64, requires_grad=False))
        self.momentum = momentum
        self.first_run = True
        self.device = device
        self.affine = affine
        if self.affine:
            self.weight = nn.Parameter(torch.ones(num_channel, dtype=torch.float32, device=self.device))
            self.bias = nn.Parameter(torch.zeros(num_channel, dtype=torch.complex64, device=self.device))
            # with torch.no_grad():
            #     self.weight.fill_(1)
            #     self.bias.fill_(0)

    def forward(self, input):

        if self.training:

            ev = torch.unsqueeze(torch.mean(input, dim=0), dim=0)

            with torch.no_grad():
                if not self.first_run:
                    self.running_mean = (ev * (1 - self.momentum) + self.running_mean * self.momentum).detach()
                else:
                    self.running_mean = ev.detach()

            var = torch.var(input, dim=0).unsqueeze(0)

            with torch.no_grad():
                if not self.first_run:
                    self.running_var = (var * (1 - self.momentum) + self.running_var * self.momentum).detach()
                else:
                    self.running_var = var.detach()
                    self.first_run = False

        else:
            var = self.running_var.detach()
            ev = self.running_mean.detach()

        input = (input - ev) / torch.sqrt(var)
        if self.affine:
            input = input * self.weight + self.bias

        return input



"""
File: create_complex_dataset.py
Author: Matthias Wolff, Florian Eilers, Xiaoyi Jiang
Description: Create a complex-valued dataset based on a lambda function. Similar to pyKAN's create_dataset
method, but complex-valued.
"""
import numpy as np
import torch


def create_complex_dataset(f, n_var, ranges, train_num=1000, test_num=1000, device="cpu"):
    """

    :param f: Lambda expression
    :param n_var: number of input variables
    :param ranges: ranges of input variables from which datapoints are randomly sampled. Possibly one seperate
    range for every input variable, or just one range for all variables combined.
    :param train_num: number of data points to be sampled as train split
    :param test_num: number of data points to be sampled as test split
    :param device: device the dataset should be created on
    :return: Dictionary of dataset (like pyKAN's create_dataset)
    """
    # if ranges is just one single tuple
    if len(np.array(ranges).shape) == 1:
        # copy this tuple to every input_variable (i.e. apply single range tuple to every input variable)
        ranges = np.array(ranges * n_var).reshape(n_var, 2)
    else:
        ranges = np.array(ranges)
    # prepare tensor (zeros) for train and test input
    train_input = torch.zeros(train_num, n_var, dtype=torch.complex64)
    test_input = torch.zeros(test_num, n_var, dtype=torch.complex64)
    for i in range(n_var):  # for every input variable i
        # sample random points within ranges[i,0] and ranges[i,1] (min max per input variable)
        train_input[:, i] = torch.rand(train_num, dtype=torch.complex64) * (ranges[i, 1] - ranges[i, 0]) + (1+1j)*ranges[i, 0]
        test_input[:, i] = torch.rand(test_num, dtype=torch.complex64) * (ranges[i, 1] - ranges[i, 0]) + (1+1j)*ranges[i, 0]

    # calculate corresponding labels / targets based on lambda expression f
    train_label = f(train_input)
    test_label = f(test_input)

    # if train labels have only 1 dimension (i.e. output is single number per datapoint)
    if len(train_label.shape) == 1:
        # then unsqueeze to Nx1 tensor
        train_label = train_label.unsqueeze(dim=1)
        test_label = test_label.unsqueeze(dim=1)


    # build dataset
    dataset = {}
    dataset['train_input'] = train_input.to(device)
    dataset['test_input'] = test_input.to(device)

    dataset['train_label'] = train_label.to(device)
    dataset['test_label'] = test_label.to(device)

    return dataset
